#include <QEventLoop>
#include <QUrl>
#include <QDir>

#include <QDebug>

#include "downloader.h"

Downloader::Downloader(PlaylistLoop *loop)
    : QObject()
    , m_loop(loop)
    , m_idl(0)
    , m_index(-1)
{
    m_baseUrl = m_loop->baseUrl();
    m_cacheDir = QDir::homePath() + "/.linutop-kiosk";
    m_dataDir = m_cacheDir + "/data";
    m_etagDir = m_cacheDir + "/etag";

    QDir d;
    d.mkpath(m_dataDir);
    d.mkpath(m_etagDir);
}

Downloader::~Downloader()
{
}

void Downloader::start()
{
    downloadNext();
}

void Downloader::downloadNext()
{
    if (m_idl)
    {
        delete m_idl;
        m_idl = 0;
    }

    ++m_index;

    if (m_index >= m_loop->count())
    {
        emit(downloadDone());
        deleteLater();
        return;
    }

    qDebug() << "DL idx:" << m_index << "/" << m_loop->count() - 1;
    PlaylistItem *item = m_loop->at(m_index);
    qDebug() << item->uri() << item->type();

    // check if we need to download the data
    if (item->type() == PlaylistItem::TypeWebpage)
    {
        downloadNext();
        return;
    }

    QString itemUri = item->uri();
    if (itemUri.startsWith("file://"))
    {
        downloadNext();
        return;
    }

    if (itemUri.startsWith("rtsp://"))
    {
        downloadNext();
        return;
    }

    if (itemUri.at(0) == '/')
        itemUri = m_baseUrl + "/r" + itemUri;

    QFileInfo info(itemUri);
    QString filename(itemUri.toUtf8().toBase64().replace("/", "-"));
    QString localPath = m_dataDir + "/" + filename;
    QString localUri = QString("file://") + localPath;
    QString etagPath = m_etagDir + "/" + filename;
    item->setUri(localUri);

    // the ItemDownloader decides if it needs to use the cache or not
    m_idl = new ItemDownloader(itemUri, localPath, etagPath);
    connect(m_idl, SIGNAL(finished()), this, SLOT(downloadNext()));
    m_idl->start();
}


ItemDownloader::ItemDownloader(QString &uri, QString &dest, QString &etag)
    : QObject()
    , m_uri(uri)
{
    m_file.setFileName(dest);
    m_etagFile.setFileName(etag);
    m_nam = new QNetworkAccessManager();
}

bool ItemDownloader::inCache()
{
    if (! m_file.exists() || ! m_etagFile.exists())
        return false;

    if (!m_etagFile.open(QIODevice::ReadOnly))
        return false;

    QString localEtag(m_etagFile.readAll());
    m_etagFile.close();

    QNetworkReply *reply = m_nam->head(m_request);

    // synchronous request
    QEventLoop loop;
    connect(reply, SIGNAL(finished()), &loop, SLOT(quit()));
    loop.exec();

    if (reply->error() != QNetworkReply::NoError)
        return true;  // try to use the cache if we couldn't HEAD the file

    QString dateStr(reply->rawHeader("Last-Modified"));
    QString length(reply->rawHeader("Content-Length"));

    QString distantEtag = QString("%1:%2").arg(dateStr, length);

    if (distantEtag != localEtag)
        return false;

    // TODO(gpocentek): check that the local file has the correct size

    return true;
}

void ItemDownloader::start()
{
    m_request = QNetworkRequest(QUrl(m_uri));

    if (inCache())
    {
        qDebug() << "NOT downloading";
        emit(finished());
        return;
    }

    qDebug() << "Downloading";
    m_reply = m_nam->get(m_request);

    connect(m_reply, SIGNAL(finished()), this, SLOT(onFinished()));
    connect(m_reply, SIGNAL(readyRead()), this, SLOT(onReadyRead()));
}

ItemDownloader::~ItemDownloader()
{
    m_nam->deleteLater();
}

void ItemDownloader::onReadyRead()
{
    if (m_reply->error() != QNetworkReply::NoError)
        return;

    if (!m_file.isOpen())
        m_file.open(QIODevice::WriteOnly);

    m_file.write(m_reply->readAll());
}

void ItemDownloader::onFinished()
{
    if (m_reply->error() != QNetworkReply::NoError)
        return;

    m_file.write(m_reply->readAll());
    m_file.close();

    QString dateStr(m_reply->rawHeader("Last-Modified"));
    QString length(m_reply->rawHeader("Content-Length"));

    QString distantEtag = QString("%1:%2").arg(dateStr, length);
    if (m_etagFile.open(QIODevice::WriteOnly))
    {
        m_etagFile.write(distantEtag.toUtf8());
        m_etagFile.close();
    }

    emit(finished());
    qDebug() << "finished download";
}
